/*******************************************************************************
  Copyright (c) 2002-2007 J. P. Abelanet/Basis Software.
  All Rights Reserved.

  -> THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF BASIS SOFTWARE. <-

  FILE
    MeshTopologyOffsetter.cpp
    
  DESCRIPTION
    Class for offsetting surface meshes.
    
  AUTHOR
    J. P. Abelanet - 05/22/02
*******************************************************************************/

/* system includes */

/* library includes */
#include <geom/Entity.h>
#include <geom/EntityList.h>
#include <mesh/MeshDeformer.h>
#include <mesh/MeshTopologyOffsetter.h>
#include <nmb/MeshModel.h>
#include <nmb/MeshVertex.h>
#include <nmb/SurfaceVertex.h>
#include <vmath/Array.h>
#include <vmath/Base.h>
#include <vmath/Error.h>
#include <vmath/Hash.h>
#include <vmath/Vector3D.h>
#include <../src/mesh/MeshGeometry.h>

/* local declarations */



namespace GE {




static Error OffsetVertex (MeshDeformer *const meshDeformer,
       MeshVertex *const vertex, const Real64 Offset, const Real64 deflectionBuffer,
       Vector3D &point)
/*==============================================================================
  ABSTRACT
    This routine offsets a single MeshVertex.
  
  PARAMETERS
    meshDeformer  INOUT
      MeshDeformer to use
    
    vertex  INOUT
      MeshVertex to reposition
    
    Offset  IN
      Distance along normals to offset

    deflectionBuffer  IN
      Normalized safety margin used to insure that the modified Edges and Faces
      do not encroach on any other part of the MeshModel - passed directly into
      MeshDeformer::Vertex_Check_Point()
      
    point  OUT
      Point that MeshVertex was attempted to be moved to, whether this succeeds
      or not
      
    returns:
      Error_self_intersection	- the deformed surface would self-intersect
      Error_degenerate_face	- one of the deformed Faces would be degenerate
      Error_nonlocal_projection	- point projection not in vicinity of starting
      				  vertex for some vertex in mesh
  
  REVISIONS
    09/21/04 - J. P. - Modified to use MeshDeformer to maintain
                       integrity of offsets
    09/21/04 - J. P. - Modified to use MeshDeformer to maintain
    02/10/05 - J. P. - Modified to use MeshDeformer class
    04/27/05 - J. P. - Modified to use EntityList for all lists of entities
    08/25/05 - J. P. - Modified to use Vector3D class
    09/16/05 - J. P. - Modified to return Point
    10/31/05 - J. P. - Added Buffer parameter to MeshDeformer::Vertex_Assign_Point
    03/13/06 - J. P. - Modified to call MESHDEF_Vertex_Check_Point()
    11/22/06 - J. P. - Removing Error return value from Topology Inquire methods
    
  AUTHOR
    J. P. Abelanet - 05/23/02
==============================================================================*/
{
  EntityList<SurfaceVertex>	SurfaceVertices;
  Vector3D	NormalSum (0.0, 0.0, 0.0);
  Int32		i;
  Error		localError;
  
  /* Get the original point */
  point = *MeshGeometry::Vertex_Inquire_Point (vertex);
  
  /* Get all the SurfaceVertices adjacent to this MeshVertex */
  vertex->Inquire_SurfaceVertices (&SurfaceVertices);

  /* Average all the normals */
  for (i = 0; i < SurfaceVertices.Size (); i++)
    NormalSum += *SurfaceVertices[i]->Inquire_Normal ();
  
  /* Compute the new point */
  point += NormalSum * Offset/(Real64)SurfaceVertices.Size ();
  
  /* Check the new point */
  if ((localError = meshDeformer->Vertex_Check_Point (
  		vertex, point, deflectionBuffer)) != Error_no_errors)
    return localError;
    
  /* Assign the new point to the MeshVertex */
  return meshDeformer->Vertex_Assign_Point (vertex, point);
} /* OffsetVertex */


Error MeshTopologyOffsetter::MeshModel_Create_Offset (const MeshModel *model, const Real64 Offset,
         const Real64 deflectionBuffer, MeshModel **const OffsetModel,
         Hash<Entity*, Entity*> **const entityHash,
	 Array<MeshTopologyOffsetter::LocalizedError> *const localizedErrors)
/*******************************************************************************
  ABSTRACT
    This routine computes the offset of a MeshModel,
    subject to the following conditions:
    - The offset location for each MeshVertex must not cause the offset MeshModel
      to transiently self-intersect, subject to the deflectionBuffer

  PARAMETERS
    model  IN
      MeshModel to offset
      
    Offset  IN
      Distance along normals to offset
      
    deflectionBuffer  IN
      Normalized safety margin used to insure that the modified Edges and Faces
      do not encroach on any other part of the MeshModel - passed directly into
      MeshDeformer::Vertex_Assign_Point()
      
    OffsetModel  OUT
      Offset of original MeshModel
      
    entityHash  INOUT
      Optional parameter - if non-NULL initially, this will contain a
      hash table that maps original topology/geometry entities to the
      corresponding copies in the new MeshModel
    
    localizedErrors  OUT
      List of localized errors if this does not return Error_no_errors
      
    returns:
      Error_self_intersection	- the offset mesh would self-intersect
      Error_degenerate_face	- one of the deformed Faces would be degenerate
      Error_nonlocal_projection	- point projection not in vicinity of starting
      				  vertex for some vertex in mesh

  BUGS
    - May need to triangulate the new MeshFaces
    - May need to recompute the MeshVertex normals
    - Does not currently generate the outgoing Hash
  
  REVISIONS
    09/21/04 - J. P. - Modified to use MeshDeformer to maintain
                       integrity of offsets
    11/28/04 - J. P. - Modified to call Entity::Copy
    02/10/05 - J. P. - Modified to use MeshDeformer class
    04/18/05 - J. P. - Modified to use EntityList for all lists of entities
    09/16/05 - J. P. - Modified to return list of localized errors
    10/31/05 - J. P. - Modified parameter list to MeshDeformer::Model_Create_MeshDeformer(),
                       and modified to take deflectionBuffer
    12/28/05 - J. P. - Modified to not compute FacePlanes
    11/16/06 - J. P. - Modified to use new Array template class
    11/22/06 - J. P. - Removing Error return value from Topology Inquire methods
    
  AUTHOR
    J. P. Abelanet - 05/22/02
*******************************************************************************/
{
  MeshTopologyOffsetter::LocalizedError	localizedError;
  MeshDeformer	*meshDeformer;
  MeshVertex	*vertex;
  Int32		NumVertices, i;
  Error		error = Error_no_errors, localError;

  /* Init the outgoing parameters */
  *OffsetModel = NULL;
  if (entityHash)
    *entityHash = NULL;
  localizedErrors->Clear ();
  
  /* Copy the MeshModel */
  if ((*OffsetModel = MeshModel::Downcast (model->Copy ())) == NULL)
    return (Error_insufficient_memory);
  
  /* Build the MeshDeformer */
  if ((localError = MeshDeformer::Model_Create_MeshDeformer (*OffsetModel,
           16 /* MaxLevel */, 8 /*MaxElPerNode*/,
	   MeshDeformer::Vertex_Grade, &meshDeformer)) != Error_no_errors)
    return (localError);
    
  /* Get the number of MeshVertices in the MeshModel */
  (*OffsetModel)->Inquire_Vertices (NULL, &NumVertices);
  
  /* Offset all the MeshVertices */
  for (i = 0; i < NumVertices; i++)
    if ((localError = meshDeformer->Inquire_NextVertex (
            &vertex)) != Error_no_errors)
      error = localError;
    else {
      localizedError.OriginalPoint = *MeshGeometry::Vertex_Inquire_Point (vertex);
      if ((localError = OffsetVertex (meshDeformer, vertex,
               Offset, deflectionBuffer, localizedError.AttemptedLocation)) != Error_no_errors) {
        error = localError;
	
	/* Add this location to the list of localized errors */
        localizedErrors->Append (localizedError);
      }
    }
    
  /* Destroy the MeshDeformer */
  delete meshDeformer;

  /* Destroy the MeshModel if any problems occurred */
  if (error != Error_no_errors) {
    (*OffsetModel)->Delete ();
    *OffsetModel = NULL;
  }
  
  return (error);
}  /* MeshTopologyOffsetter::MeshModel_Create_Offset */


}	// namespace GE
